"""Scopes in which a W&B Automation can be triggered."""

from __future__ import annotations

from typing import Literal, Union

from pydantic import BeforeValidator, Field
from typing_extensions import Annotated, TypeAlias, get_args

from wandb._pydantic import GQLBase
from wandb.automations._generated import (
    ArtifactPortfolioScopeFields,
    ArtifactSequenceScopeFields,
    ProjectScopeFields,
)

from ._generated import TriggerScopeType
from ._validators import validate_scope

# NOTE: Re-defined publicly with a more readable name for easier access
ScopeType = TriggerScopeType
"""The type of scope that triggers an automation."""


class _BaseScope(GQLBase):
    scope_type: ScopeType


class _ArtifactSequenceScope(_BaseScope, ArtifactSequenceScopeFields):
    """The ID and name of a "sequence"-type ArtifactCollection scope of an automation."""

    scope_type: Literal[ScopeType.ARTIFACT_COLLECTION] = ScopeType.ARTIFACT_COLLECTION


class _ArtifactPortfolioScope(_BaseScope, ArtifactPortfolioScopeFields):
    """The ID and name of a "portfolio"-type ArtifactCollection scope of an automation."""

    scope_type: Literal[ScopeType.ARTIFACT_COLLECTION] = ScopeType.ARTIFACT_COLLECTION


# for type annotations
ArtifactCollectionScope = Annotated[
    Union[_ArtifactSequenceScope, _ArtifactPortfolioScope],
    BeforeValidator(validate_scope),
    Field(discriminator="typename__"),
]
"""The ID and name of the ArtifactCollection scope of an automation."""

# for runtime type checks
ArtifactCollectionScopeTypes: tuple[type, ...] = (
    _ArtifactSequenceScope,
    _ArtifactPortfolioScope,
)


class ProjectScope(_BaseScope, ProjectScopeFields):
    """The ID and name of the Project scope of an automation."""

    scope_type: Literal[ScopeType.PROJECT] = ScopeType.PROJECT


# for type annotations
AutomationScope: TypeAlias = Annotated[
    Union[_ArtifactSequenceScope, _ArtifactPortfolioScope, ProjectScope],
    BeforeValidator(validate_scope),
    Field(discriminator="typename__"),
]
# for runtime type checks
AutomationScopeTypes: tuple[type, ...] = get_args(AutomationScope)

# Aliases for naming clarity/consistency
SavedScope: TypeAlias = AutomationScope
InputScope: TypeAlias = AutomationScope

SavedScopeTypes: tuple[type, ...] = get_args(SavedScope)
InputScopeTypes: tuple[type, ...] = get_args(InputScope)



from __future__ import annotations

from functools import singledispatch
from itertools import chain
from typing import Any, TypeVar

from ._filters import And, FilterExpr, In, Nor, Not, NotIn, Op, Or

T = TypeVar("T")


def validate_scope(v: Any) -> Any:
    """Convert a familiar wandb `Project` or `ArtifactCollection` object to an automation scope."""
    from wandb.apis.public import ArtifactCollection, Project

    from .scopes import ProjectScope, _ArtifactPortfolioScope, _ArtifactSequenceScope

    if isinstance(v, Project):
        return ProjectScope(id=v.id, name=v.name)
    if isinstance(v, ArtifactCollection):
        cls = _ArtifactSequenceScope if v.is_sequence() else _ArtifactPortfolioScope
        return cls(id=v.id, name=v.name)
    return v


@singledispatch
def simplify_op(op: Op | FilterExpr) -> Op | FilterExpr:
    """Simplify a MongoDB filter by removing and unnesting redundant operators."""
    return op


@simplify_op.register
def _(op: And) -> Op:
    # {"$and": []} -> {"$and": []}
    if not (args := op.and_):
        return op

    # {"$and": [op]} -> op
    if len(args) == 1:
        return simplify_op(args[0])

    # {"$and": [op, {"$and": [op2, ...]}]} -> {"$and": [op, op2, ...]}
    flattened = chain.from_iterable(x.and_ if isinstance(x, And) else [x] for x in args)
    return And(and_=map(simplify_op, flattened))


@simplify_op.register
def _(op: Or) -> Op:
    # {"$or": []} -> {"$or": []}
    if not (args := op.or_):
        return op

    # {"$or": [op]} -> op
    if len(args) == 1:
        return simplify_op(args[0])

    # {"$or": [op, {"$or": [op2, ...]}]} -> {"$or": [op, op2, ...]}
    flattened = chain.from_iterable(x.or_ if isinstance(x, Or) else [x] for x in args)
    return Or(or_=map(simplify_op, flattened))


@simplify_op.register
def _(op: Not) -> Op:
    inner = op.not_

    # {"$not": {"$not": op}} -> op
    # {"$not": {"$or": [op, ...]}} -> {"$nor": [op, ...]}
    # {"$not": {"$nor": [op, ...]}} -> {"$or": [op, ...]}
    # {"$not": {"$in": [op, ...]}} -> {"$nin": [op, ...]}
    # {"$not": {"$nin": [op, ...]}} -> {"$in": [op, ...]}
    if isinstance(inner, (Not, Or, Nor, In, NotIn)):
        return simplify_op(~inner)
    return Not(not_=simplify_op(inner))



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import DeleteTriggerResult


class DeleteTrigger(GQLBase):
    delete_trigger: Optional[DeleteTriggerResult] = Field(alias="deleteTrigger")


DeleteTrigger.model_rebuild()



from wandb._pydantic import IS_PYDANTIC_V2

from . import _filters as filters
from . import actions, automations, events, scopes
from .actions import ActionType, DoNothing, DoNotification, DoWebhook
from .automations import Automation, NewAutomation, PreparedAutomation
from .events import (
    ArtifactEvent,
    EventType,
    OnAddArtifactAlias,
    OnCreateArtifact,
    OnLinkArtifact,
    OnRunMetric,
    RunEvent,
)
from .integrations import Integration, SlackIntegration, WebhookIntegration
from .scopes import ArtifactCollectionScope, ProjectScope, ScopeType

# ----------------------------------------------------------------------------
# WARNINGS on import
if not IS_PYDANTIC_V2:
    # Raises an error in Pydantic v1 environments, where the Automations API
    # has not been tested and is unlikely to work as expected.
    #
    # Remove this when we either:
    # - Drop support for Pydantic v1
    # - Are able to implement (limited) Pydantic v1 support
    raise ImportError(
        "The W&B Automations API is not supported in Pydantic v1 at this time. "
        "If at all possible, we currently recommend upgrading to Pydantic v2 to use this feature.",
    )

else:
    # If Pydantic v2 is available, we can use the full Automations API
    # but communicate to users that the API is still experimental and
    # may change rapidly.
    import warnings

    warnings.warn(
        "The W&B Automations API is currently experimental. Although we'll communicate "
        "breaking changes in release notes and attempt to minimize them in general, "
        "please know that such changes may occur between release versions without notice. "
        "We strongly recommend pinning your `wandb` version when using the Automations API "
        "to avoid unexpected breakages.",
        FutureWarning,
        stacklevel=1,
    )
# ----------------------------------------------------------------------------

__all__ = [
    "filters",
    "scopes",
    "events",
    "actions",
    "automations",
    # Scopes
    "ScopeType",
    "ArtifactCollectionScope",
    "ProjectScope",
    # Events
    "EventType",
    "OnAddArtifactAlias",
    "OnCreateArtifact",
    "OnLinkArtifact",
    "OnRunMetric",
    "ArtifactEvent",
    "RunEvent",
    # Actions
    "ActionType",
    "DoNotification",
    "DoWebhook",
    "DoNothing",
    # Automations
    "Automation",
    "NewAutomation",
    "PreparedAutomation",
    # Integrations
    "Integration",
    "SlackIntegration",
    "WebhookIntegration",
]



"""Actions that are triggered by W&B Automations."""

# ruff: noqa: UP007  # Avoid using `X | Y` for union fields, as this can cause issues with pydantic < 2.6

from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import Field
from typing_extensions import Self, TypeAlias, get_args

from wandb._pydantic import (
    SerializedToJson,
    field_validator,
    model_validator,
    pydantic_isinstance,
)

from ._generated import (
    AlertSeverity,
    GenericWebhookActionFields,
    GenericWebhookActionInput,
    NoOpActionFields,
    NoOpTriggeredActionInput,
    NotificationActionFields,
    NotificationActionInput,
    QueueJobActionFields,
    TriggeredActionType,
)
from .integrations import SlackIntegration, WebhookIntegration

# Note: Pydantic doesn't like `list['JsonValue']` or `dict[str, 'JsonValue']`,
# which causes a RecursionError.
JsonValue: TypeAlias = Union[
    List[Any],
    Dict[str, Any],
    # NOTE: For now, we're not expecting any doubly-serialized strings, as this makes validation logic easier, but revisit and revise if needed.
    # str,
    bool,
    int,
    float,
    None,
]

# NOTE: Name shortened for readability and defined publicly for easier access
ActionType = TriggeredActionType
"""The type of action triggered by an automation."""


# ------------------------------------------------------------------------------
# Saved types: for parsing response data from saved automations


# NOTE: `QueueJobActionInput` for defining a Launch job is deprecated,
# so while we allow parsing it from previously saved Automations, we deliberately
# don't currently expose it in the API for creating automations.
class SavedLaunchJobAction(QueueJobActionFields):
    action_type: Literal[ActionType.QUEUE_JOB] = ActionType.QUEUE_JOB


class SavedNotificationAction(NotificationActionFields):
    action_type: Literal[ActionType.NOTIFICATION] = ActionType.NOTIFICATION


class SavedWebhookAction(GenericWebhookActionFields):
    action_type: Literal[ActionType.GENERIC_WEBHOOK] = ActionType.GENERIC_WEBHOOK

    # We override the type of the `requestPayload` field since the original GraphQL
    # schema (and generated class) effectively defines it as a string, when we know
    # and need to anticipate the expected structure of the JSON-serialized data.
    request_payload: Optional[SerializedToJson[JsonValue]] = Field(  # type: ignore[assignment]
        default=None, alias="requestPayload"
    )


class SavedNoOpAction(NoOpActionFields):
    action_type: Literal[ActionType.NO_OP] = ActionType.NO_OP


# for type annotations
SavedAction = Union[
    SavedLaunchJobAction,
    SavedNotificationAction,
    SavedWebhookAction,
    SavedNoOpAction,
]
# for runtime type checks
SavedActionTypes: tuple[type, ...] = get_args(SavedAction)


# ------------------------------------------------------------------------------
# Input types: for creating or updating automations
class DoNotification(NotificationActionInput):
    """Schema for defining a triggered notification action."""

    action_type: Literal[ActionType.NOTIFICATION] = ActionType.NOTIFICATION

    # Note: Validation aliases match arg names from `wandb.alert()` to allow
    # continuity with previous API.
    title: str = Field(default="", validation_alias="title")
    message: str = Field(default="", validation_alias="text")
    severity: AlertSeverity = Field(
        default=AlertSeverity.INFO, validation_alias="level"
    )

    @field_validator("severity", mode="before")
    @classmethod
    def _validate_severity(cls, v: Any) -> Any:
        # Be helpful by accepting case-insensitive strings
        return v.upper() if isinstance(v, str) else v

    @model_validator(mode="before")
    @classmethod
    def _from_saved(cls, v: Any) -> Any:
        """Convert an action on a saved automation to a new/input action."""
        if pydantic_isinstance(v, SavedNotificationAction):
            return cls(
                integration_id=v.integration.id,
                title=v.title,
                message=v.message,
                severity=v.severity,
            )
        return v

    @classmethod
    def from_integration(
        cls,
        integration: SlackIntegration,
        *,
        title: str = "",
        text: str = "",
        level: AlertSeverity = AlertSeverity.INFO,
    ) -> Self:
        """Define a notification action that sends to the given (Slack) integration."""
        integration = SlackIntegration.model_validate(integration)
        return cls(
            integration_id=integration.id,
            title=title,
            message=text,
            severity=level,
        )


class DoWebhook(GenericWebhookActionInput):
    """Schema for defining a triggered webhook action."""

    action_type: Literal[ActionType.GENERIC_WEBHOOK] = ActionType.GENERIC_WEBHOOK

    # overrides the generated field type to parse/serialize JSON strings
    request_payload: Optional[SerializedToJson[JsonValue]] = Field(  # type: ignore[assignment]
        default=None, alias="requestPayload"
    )

    @model_validator(mode="before")
    @classmethod
    def _from_saved(cls, v: Any) -> Any:
        """Convert an action on a saved automation to a new/input action."""
        if pydantic_isinstance(v, SavedWebhookAction):
            return cls(
                integration_id=v.integration.id,
                request_payload=v.request_payload,
            )
        return v

    @classmethod
    def from_integration(
        cls,
        integration: WebhookIntegration,
        *,
        request_payload: Optional[SerializedToJson[JsonValue]] = None,
    ) -> Self:
        """Define a webhook action that sends to the given (webhook) integration."""
        integration = WebhookIntegration.model_validate(integration)
        return cls(integration_id=integration.id, request_payload=request_payload)


class DoNothing(NoOpTriggeredActionInput):
    """Schema for defining a triggered no-op action."""

    action_type: Literal[ActionType.NO_OP] = ActionType.NO_OP

    no_op: bool = True  # prevent exclusion on `.model_dump(exclude_none=True)`

    @field_validator("no_op", mode="before")
    @classmethod
    def _ensure_nonnull(cls, v: Any) -> Any:
        # Ensuring the value isn't None complies with validation and
        # prevents the field (and action data) from getting excluded on
        # `.model_dump(exclude_none=True)`
        return True if (v is None) else v


DoNotification.model_rebuild()
DoWebhook.model_rebuild()
DoNothing.model_rebuild()


# for type annotations
InputAction = Union[
    DoNotification,
    DoWebhook,
    DoNothing,
]
# for runtime type checks
InputActionTypes: tuple[type, ...] = get_args(InputAction)



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import CreateFilterTriggerResult


class CreateFilterTrigger(GQLBase):
    create_filter_trigger: Optional[CreateFilterTriggerResult] = Field(
        alias="createFilterTrigger"
    )


CreateFilterTrigger.model_rebuild()



"""Reliably launch and connect to backend server process (wandb service).

Backend server process can be connected to using tcp sockets transport.
"""

import datetime
import os
import pathlib
import platform
import shutil
import subprocess
import sys
import tempfile
import time
from typing import TYPE_CHECKING, Any, Dict, Optional

from wandb import _sentry
from wandb.env import core_debug, error_reporting_enabled, is_require_legacy_service
from wandb.errors import Error, WandbCoreNotAvailableError
from wandb.errors.term import termlog, termwarn
from wandb.util import get_core_path, get_module

from . import _startup_debug, port_file

if TYPE_CHECKING:
    from wandb.sdk.wandb_settings import Settings


class ServiceStartProcessError(Error):
    """Raised when a known error occurs when launching wandb service."""


class ServiceStartTimeoutError(Error):
    """Raised when service start times out."""


class ServiceStartPortError(Error):
    """Raised when service start fails to find a port."""


class _Service:
    _settings: "Settings"
    _sock_port: Optional[int]
    _internal_proc: Optional[subprocess.Popen]
    _startup_debug_enabled: bool

    def __init__(
        self,
        settings: "Settings",
    ) -> None:
        self._settings = settings
        self._stub = None
        self._sock_port = None
        self._internal_proc = None
        self._startup_debug_enabled = _startup_debug.is_enabled()

        _sentry.configure_scope(tags=dict(settings), process_context="service")

    def _startup_debug_print(self, message: str) -> None:
        if not self._startup_debug_enabled:
            return
        _startup_debug.print_message(message)

    def _wait_for_ports(
        self, fname: str, proc: Optional[subprocess.Popen] = None
    ) -> None:
        """Wait for the service to write the port file and then read it.

        Args:
            fname: The path to the port file.
            proc: The process to wait for.

        Raises:
            ServiceStartTimeoutError: If the service takes too long to start.
            ServiceStartPortError: If the service writes an invalid port file or unable to read it.
            ServiceStartProcessError: If the service process exits unexpectedly.

        """
        time_max = time.monotonic() + self._settings.x_service_wait
        while time.monotonic() < time_max:
            if proc and proc.poll():
                # process finished
                # define these variables for sentry context grab:
                # command = proc.args
                # sys_executable = sys.executable
                # which_python = shutil.which("python3")
                # proc_out = proc.stdout.read()
                # proc_err = proc.stderr.read()
                context = dict(
                    command=proc.args,
                    sys_executable=sys.executable,
                    which_python=shutil.which("python3"),
                    proc_out=proc.stdout.read() if proc.stdout else "",
                    proc_err=proc.stderr.read() if proc.stderr else "",
                )
                raise ServiceStartProcessError(
                    f"The wandb service process exited with {proc.returncode}. "
                    "Ensure that `sys.executable` is a valid python interpreter. "
                    "You can override it with the `_executable` setting "
                    "or with the `WANDB_X_EXECUTABLE` environment variable."
                    f"\n{context}",
                    context=context,
                )
            if not os.path.isfile(fname):
                time.sleep(0.2)
                continue
            try:
                pf = port_file.PortFile()
                pf.read(fname)
                if not pf.is_valid:
                    time.sleep(0.2)
                    continue
                self._sock_port = pf.sock_port
            except Exception as e:
                # todo: point at the docs. this could be due to a number of reasons,
                #  for example, being unable to write to the port file etc.
                raise ServiceStartPortError(
                    f"Failed to allocate port for wandb service: {e}."
                )
            return
        raise ServiceStartTimeoutError(
            "Timed out waiting for wandb service to start after "
            f"{self._settings.x_service_wait} seconds. "
            "Try increasing the timeout with the `_service_wait` setting."
        )

    def _launch_server(self) -> None:
        """Launch server and set ports."""
        # References for starting processes
        # - https://github.com/wandb/wandb/blob/archive/old-cli/wandb/__init__.py
        # - https://stackoverflow.com/questions/1196074/how-to-start-a-background-process-in-python
        self._startup_debug_print("launch")

        kwargs: Dict[str, Any] = dict(close_fds=True)
        # flags to handle keyboard interrupt signal that is causing a hang
        if platform.system() == "Windows":
            kwargs.update(creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)  # type: ignore [attr-defined]
        else:
            kwargs.update(start_new_session=True)

        pid = str(os.getpid())

        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, f"port-{pid}.txt")

            executable = self._settings.x_executable
            exec_cmd_list = [executable, "-m"]

            service_args = []

            if not is_require_legacy_service():
                try:
                    core_path = get_core_path()
                except WandbCoreNotAvailableError as e:
                    _sentry.reraise(e)

                service_args.extend([core_path])

                if not error_reporting_enabled():
                    service_args.append("--no-observability")

                if core_debug(default="False"):
                    service_args.extend(["--log-level", "-4"])

                exec_cmd_list = []
            else:
                service_args.extend(["wandb", "service", "--debug"])
                termwarn(
                    "Using legacy-service, which is deprecated. If this is"
                    " unintentional, you can fix it by ensuring you do not call"
                    " `wandb.require('legacy-service')` and do not set the"
                    " WANDB_X_REQUIRE_LEGACY_SERVICE environment"
                    " variable."
                )

            service_args += [
                "--port-filename",
                fname,
                "--pid",
                pid,
            ]

            if os.environ.get("WANDB_SERVICE_PROFILE") == "memray":
                _ = get_module(
                    "memray",
                    required=(
                        "wandb service memory profiling requires memray, "
                        "install with `pip install memray`"
                    ),
                )

                time_tag = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
                output_file = f"wandb_service.memray.{time_tag}.bin"
                cli_executable = (
                    pathlib.Path(__file__).parent.parent.parent.parent
                    / "tools"
                    / "cli.py"
                )
                exec_cmd_list = [
                    executable,
                    "-m",
                    "memray",
                    "run",
                    "-o",
                    output_file,
                ]
                service_args[0] = str(cli_executable)
                termlog(
                    f"wandb service memory profiling enabled, output file: {output_file}"
                )
                termlog(
                    f"Convert to flamegraph with: `python -m memray flamegraph {output_file}`"
                )

            try:
                internal_proc = subprocess.Popen(
                    exec_cmd_list + service_args,  # type: ignore[arg-type]
                    env=os.environ,
                    **kwargs,
                )
            except Exception as e:
                _sentry.reraise(e)

            self._startup_debug_print("wait_ports")
            try:
                self._wait_for_ports(fname, proc=internal_proc)
            except Exception as e:
                _sentry.reraise(e)
            self._startup_debug_print("wait_ports_done")
            self._internal_proc = internal_proc
        self._startup_debug_print("launch_done")

    def start(self) -> None:
        self._launch_server()

    @property
    def sock_port(self) -> Optional[int]:
        return self._sock_port

    def join(self) -> int:
        ret = 0
        if self._internal_proc:
            ret = self._internal_proc.wait()
        return ret



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import UpdateFilterTriggerResult


class UpdateFilterTrigger(GQLBase):
    update_filter_trigger: Optional[UpdateFilterTriggerResult] = Field(
        alias="updateFilterTrigger"
    )


UpdateFilterTrigger.model_rebuild()



# ruff: noqa: UP007

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Final, Literal, Optional, Union, overload

from pydantic import Field, PositiveInt, StrictFloat, StrictInt, field_validator
from typing_extensions import Self, override

from wandb._pydantic.base import Base, GQLBase

from .expressions import FilterExpr
from .operators import BaseOp, RichReprResult

if TYPE_CHECKING:
    from wandb.automations.events import RunMetricFilter

# Maps MongoDB comparison operators -> Python literal (str) representations
MONGO2PY_OPS: Final[dict[str, str]] = {
    "$eq": "==",
    "$ne": "!=",
    "$gt": ">",
    "$lt": "<",
    "$gte": ">=",
    "$lte": "<=",
}
# Reverse mapping from Python literal (str) -> MongoDB operator key
PY2MONGO_OPS: Final[dict[str, str]] = {v: k for k, v in MONGO2PY_OPS.items()}


class Agg(str, Enum):  # from `Aggregation`
    """Supported run metric aggregation operations."""

    MAX = "MAX"
    MIN = "MIN"
    AVERAGE = "AVERAGE"


class ChangeType(str, Enum):  # from `RunMetricChangeType`
    """Describes the metric change as absolute (arithmetic difference) or relative (decimal percentage)."""

    ABSOLUTE = "ABSOLUTE"
    RELATIVE = "RELATIVE"


class ChangeDirection(str, Enum):  # from `RunMetricChangeDirection`
    """Describes the direction of the metric change."""

    INCREASE = "INCREASE"
    DECREASE = "DECREASE"
    ANY = "ANY"


class _BaseMetricFilter(GQLBase):
    name: str
    """Name of the observed metric."""

    agg: Optional[Agg]
    """Aggregation operation, if any, to apply over the window size."""

    window: PositiveInt
    """Size of the window over which the metric is aggregated."""

    # ------------------------------------------------------------------------------

    threshold: Union[StrictInt, StrictFloat]
    """Threshold value to compare against."""

    @field_validator("agg", mode="before")
    @classmethod
    def _validate_agg(cls, v: Any) -> Any:
        # Be helpful: e.g. "min" -> "MIN"
        return v.strip().upper() if isinstance(v, str) else v

    @overload
    def __and__(self, other: BaseOp | FilterExpr) -> RunMetricFilter: ...
    @overload
    def __and__(self, other: Any) -> Any: ...
    def __and__(self, other: BaseOp | FilterExpr | Any) -> RunMetricFilter | Any:
        """Supports syntactic sugar for defining a triggering RunMetricEvent from `run_metric_filter & run_filter`."""
        from wandb.automations.events import RunMetricFilter, _InnerRunMetricFilter

        if isinstance(run_filter := other, (BaseOp, FilterExpr)):
            # Assume `other` is a run filter, and we are building a RunMetricEvent.
            # For the metric filter, delegate to the inner validator(s) to further wrap/nest as appropriate.
            metric_filter = _InnerRunMetricFilter.model_validate(self)
            return RunMetricFilter(
                run_metric_filter=metric_filter, run_filter=run_filter
            )
        return other.__and__(self)  # Try switching the order of operands


class MetricThresholdFilter(_BaseMetricFilter):  # from `RunMetricThresholdFilter`
    """For run events, defines a metric filter comparing a metric against a user-defined threshold value."""

    name: str
    agg: Optional[Agg] = Field(default=None, alias="agg_op")
    window: PositiveInt = Field(default=1, alias="window_size")

    cmp: Literal["$gte", "$gt", "$lt", "$lte"] = Field(alias="cmp_op")
    """Comparison operator used to compare the metric value (left) vs. the threshold value (right)."""

    threshold: Union[StrictInt, StrictFloat]

    @field_validator("cmp", mode="before")
    @classmethod
    def _validate_cmp(cls, v: Any) -> Any:
        # Be helpful: e.g. ">" -> "$gt"
        return PY2MONGO_OPS.get(v.strip(), v) if isinstance(v, str) else v

    def __repr__(self) -> str:
        metric = f"{self.agg.value}({self.name})" if self.agg else self.name
        op = MONGO2PY_OPS.get(self.cmp, self.cmp)
        expr = rf"{metric} {op} {self.threshold}"
        return repr(expr)

    @override
    def __rich_repr__(self) -> RichReprResult:  # type: ignore[override]
        yield None, repr(self)


class MetricChangeFilter(_BaseMetricFilter):  # from `RunMetricChangeFilter`
    # FIXME:
    # - `prior_window` should be optional and default to `window` if not provided.
    # - implement declarative syntax for `MetricChangeFilter` similar to `MetricThresholdFilter`.
    # - split this into tagged union of relative/absolute change filters.

    name: str
    agg: Optional[Agg] = Field(default=None, alias="agg_op")

    # FIXME: Set the `prior_window` to `window` if it's not provided, for convenience.
    window: PositiveInt = Field(alias="current_window_size")
    prior_window: PositiveInt = Field(alias="prior_window_size")
    """Size of the preceding window over which the metric is aggregated."""

    # NOTE: `cmp_op` isn't a field here.  In the backend, it's effectively `cmp_op` = "$gte"

    change_type: ChangeType = Field(alias="change_type")
    change_direction: ChangeDirection = Field(alias="change_dir")

    threshold: Union[StrictInt, StrictFloat] = Field(alias="change_amount")


class MetricOperand(Base):
    name: str
    agg: Optional[Agg] = Field(default=None, alias="agg_op")
    window: PositiveInt = Field(default=1, alias="window_size")

    def _agg(self, op: Agg, window: int) -> Self:
        if self.agg is None:  # Prevent overwriting an existing aggregation operator
            return self.model_copy(update={"agg": op, "window": window})
        raise ValueError(f"Aggregation operator already set as: {self.agg!r}")

    def max(self, window: int) -> Self:
        return self._agg(Agg.MAX, window)

    def min(self, window: int) -> Self:
        return self._agg(Agg.MIN, window)

    def average(self, window: int) -> Self:
        return self._agg(Agg.AVERAGE, window)

    # Aliased method for users familiar with e.g. torch/tf/numpy/pandas/polars/etc.
    def mean(self, window: int) -> Self:
        return self.average(window=window)

    def gt(self, other: int | float) -> MetricThresholdFilter:
        return MetricThresholdFilter(**dict(self), cmp="$gt", threshold=other)

    def lt(self, other: int | float) -> MetricThresholdFilter:
        return MetricThresholdFilter(**dict(self), cmp="$lt", threshold=other)

    def gte(self, other: int | float) -> MetricThresholdFilter:
        return MetricThresholdFilter(**dict(self), cmp="$gte", threshold=other)

    def lte(self, other: int | float) -> MetricThresholdFilter:
        return MetricThresholdFilter(**dict(self), cmp="$lte", threshold=other)

    __gt__ = gt
    __lt__ = lt
    __ge__ = gte
    __le__ = lte



"""Public API: registries."""

import json
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Sequence

if TYPE_CHECKING:
    from wandb_gql import Client

from wandb_gql import gql

import wandb
from wandb.apis.paginator import Paginator
from wandb.apis.public.artifacts import ArtifactCollection
from wandb.sdk.artifacts._graphql_fragments import (
    _gql_artifact_fragment,
    _gql_registry_fragment,
)
from wandb.sdk.artifacts._validators import REGISTRY_PREFIX


class Registries(Paginator):
    """Iterator that returns Registries."""

    QUERY = gql(
        """
        query Registries($organization: String!, $filters: JSONString, $cursor: String, $perPage: Int) {
            organization(name: $organization) {
                orgEntity {
                    name
                    projects(filters: $filters, after: $cursor, first: $perPage) {
                        pageInfo {
                            endCursor
                            hasNextPage
                        }
                        edges {
                            node {
                                ...RegistryFragment
                            }
                        }
                    }
                }
            }
        }
        """
        + _gql_registry_fragment()
    )

    def __init__(
        self,
        client: "Client",
        organization: str,
        filter: Optional[Dict[str, Any]] = None,
        per_page: Optional[int] = 100,
    ):
        self.client = client
        self.organization = organization
        self.filter = _ensure_registry_prefix_on_names(filter or {})
        variables = {
            "organization": organization,
            "filters": json.dumps(self.filter),
        }

        super().__init__(client, variables, per_page)

    def __bool__(self):
        return len(self) > 0 or len(self.objects) > 0

    def __next__(self):
        # Implement custom next since its possible to load empty pages because of auth
        self.index += 1
        while len(self.objects) <= self.index:
            if not self._load_page():
                raise StopIteration
        return self.objects[self.index]

    def collections(self, filter: Optional[Dict[str, Any]] = None) -> "Collections":
        return Collections(
            self.client,
            self.organization,
            registry_filter=self.filter,
            collection_filter=filter,
        )

    def versions(self, filter: Optional[Dict[str, Any]] = None) -> "Versions":
        return Versions(
            self.client,
            self.organization,
            registry_filter=self.filter,
            collection_filter=None,
            artifact_filter=filter,
        )

    @property
    def length(self):
        if self.last_response:
            return len(
                self.last_response["organization"]["orgEntity"]["projects"]["edges"]
            )
        else:
            return None

    @property
    def more(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"]["projects"][
                "pageInfo"
            ]["hasNextPage"]
        else:
            return True

    @property
    def cursor(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"]["projects"][
                "pageInfo"
            ]["endCursor"]
        else:
            return None

    def convert_objects(self):
        if not self.last_response:
            return []
        if (
            not self.last_response["organization"]
            or not self.last_response["organization"]["orgEntity"]
        ):
            raise ValueError(
                f"Organization '{self.organization}' not found. Please verify the organization name is correct"
            )

        return [
            Registry(
                self.client,
                self.organization,
                self.last_response["organization"]["orgEntity"]["name"],
                r["node"]["name"],
                r["node"],
            )
            for r in self.last_response["organization"]["orgEntity"]["projects"][
                "edges"
            ]
        ]


class Registry:
    """A single registry in the Registry."""

    def __init__(
        self,
        client: "Client",
        organization: str,
        entity: str,
        full_name: str,
        attrs: Dict[str, Any],
    ):
        self.client = client
        self._full_name = full_name
        self._name = full_name.replace(REGISTRY_PREFIX, "")
        self._entity = entity
        self._organization = organization
        self._description = attrs.get("description", "")
        self._allow_all_artifact_types = attrs.get(
            "allowAllArtifactTypesInRegistry", False
        )
        self._artifact_types = [
            t["node"]["name"] for t in attrs.get("artifactTypes", {}).get("edges", [])
        ]
        self._id = attrs.get("id", "")
        self._created_at = attrs.get("createdAt", "")
        self._updated_at = attrs.get("updatedAt", "")

    @property
    def full_name(self):
        return self._full_name

    @property
    def name(self):
        return self._name

    @property
    def entity(self):
        return self._entity

    @property
    def organization(self):
        return self._organization

    @property
    def description(self):
        return self._description

    @property
    def allow_all_artifact_types(self):
        return self._allow_all_artifact_types

    @property
    def artifact_types(self):
        return self._artifact_types

    @property
    def created_at(self):
        return self._created_at

    @property
    def updated_at(self):
        return self._updated_at

    @property
    def path(self):
        return [self.entity, self.name]

    def collections(self, filter: Optional[Dict[str, Any]] = None):
        registry_filter = {
            "name": self.full_name,
        }
        return Collections(self.client, self.organization, registry_filter, filter)

    def versions(self, filter: Optional[Dict[str, Any]] = None):
        registry_filter = {
            "name": self.full_name,
        }
        return Versions(self.client, self.organization, registry_filter, None, filter)


class Collections(Paginator):
    """Iterator that returns Artifact collections in the Registry."""

    QUERY = gql(
        """
        query Collections(
            $organization: String!,
            $registryFilter: JSONString,
            $collectionFilter: JSONString,
            $collectionTypes: [ArtifactCollectionType!],
            $cursor: String,
            $perPage: Int
        ) {
            organization(name: $organization) {
                orgEntity {
                    name
                    artifactCollections(
                        projectFilters: $registryFilter,
                        filters: $collectionFilter,
                        collectionTypes: $collectionTypes,
                        after: $cursor,
                        first: $perPage
                    ) {
                        totalCount
                        pageInfo {
                            endCursor
                            hasNextPage
                        }
                        edges {
                            cursor
                            node {
                                id
                                name
                                description
                                createdAt
                                tags {
                                    edges {
                                        node {
                                            name
                                        }
                                    }
                                }
                                project {
                                    name
                                    entity {
                                        name
                                    }
                                }
                                defaultArtifactType {
                                    name
                                }
                                aliases {
                                    edges {
                                        node {
                                            alias
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        """
    )

    def __init__(
        self,
        client: "Client",
        organization: str,
        registry_filter: Optional[Dict[str, Any]] = None,
        collection_filter: Optional[Dict[str, Any]] = None,
        per_page: Optional[int] = 100,
    ):
        self.client = client
        self.organization = organization
        self.registry_filter = registry_filter
        self.collection_filter = collection_filter or {}

        variables = {
            "registryFilter": json.dumps(self.registry_filter)
            if self.registry_filter
            else None,
            "collectionFilter": json.dumps(self.collection_filter)
            if self.collection_filter
            else None,
            "organization": self.organization,
            "collectionTypes": ["PORTFOLIO"],
            "perPage": per_page,
        }

        super().__init__(client, variables, per_page)

    def __bool__(self):
        return len(self) > 0 or len(self.objects) > 0

    def __next__(self):
        # Implement custom next since its possible to load empty pages because of auth
        self.index += 1
        while len(self.objects) <= self.index:
            if not self._load_page():
                raise StopIteration
        return self.objects[self.index]

    def versions(self, filter: Optional[Dict[str, Any]] = None) -> "Versions":
        return Versions(
            self.client,
            self.organization,
            registry_filter=self.registry_filter,
            collection_filter=self.collection_filter,
            artifact_filter=filter,
        )

    @property
    def length(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["totalCount"]
        else:
            return None

    @property
    def more(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["pageInfo"]["hasNextPage"]
        else:
            return True

    @property
    def cursor(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["pageInfo"]["endCursor"]
        else:
            return None

    def convert_objects(self):
        if not self.last_response:
            return []
        if (
            not self.last_response["organization"]
            or not self.last_response["organization"]["orgEntity"]
        ):
            raise ValueError(
                f"Organization '{self.organization}' not found. Please verify the organization name is correct"
            )

        return [
            ArtifactCollection(
                self.client,
                r["node"]["project"]["entity"]["name"],
                r["node"]["project"]["name"],
                r["node"]["name"],
                r["node"]["defaultArtifactType"]["name"],
                self.organization,
                r["node"],
            )
            for r in self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["edges"]
        ]


class Versions(Paginator):
    """Iterator that returns Artifact versions in the Registry."""

    def __init__(
        self,
        client: "Client",
        organization: str,
        registry_filter: Optional[Dict[str, Any]] = None,
        collection_filter: Optional[Dict[str, Any]] = None,
        artifact_filter: Optional[Dict[str, Any]] = None,
        per_page: int = 100,
    ):
        self.client = client
        self.organization = organization
        self.registry_filter = registry_filter
        self.collection_filter = collection_filter
        self.artifact_filter = artifact_filter or {}
        self.QUERY = gql(
            """
            query Versions(
                $organization: String!,
                $registryFilter: JSONString,
                $collectionFilter: JSONString,
                $artifactFilter: JSONString,
                $cursor: String,
                $perPage: Int
            ) {
                organization(name: $organization) {
                    orgEntity {
                        name
                        artifactMemberships(
                            projectFilters: $registryFilter,
                            collectionFilters: $collectionFilter,
                            filters: $artifactFilter,
                            after: $cursor,
                            first: $perPage
                        ) {
                            pageInfo {
                                endCursor
                                hasNextPage
                            }
                            edges {
                                node {
                                    artifactCollection {
                                        project {
                                            name
                                            entity {
                                                name
                                            }
                                        }
                                        name
                                    }
                                    versionIndex
                                    artifact {
                                        ...ArtifactFragment
                                    }
                                    aliases {
                                        alias
                                    }
                                }
                            }
                        }
                    }
                }
            }
            """
            + _gql_artifact_fragment(include_aliases=False)
        )

        variables = {
            "registryFilter": json.dumps(self.registry_filter)
            if self.registry_filter
            else None,
            "collectionFilter": json.dumps(self.collection_filter)
            if self.collection_filter
            else None,
            "artifactFilter": json.dumps(self.artifact_filter)
            if self.artifact_filter
            else None,
            "organization": self.organization,
        }

        super().__init__(client, variables, per_page)

    def __next__(self):
        # Implement custom next since its possible to load empty pages because of auth
        self.index += 1
        while len(self.objects) <= self.index:
            if not self._load_page():
                raise StopIteration
        return self.objects[self.index]

    def __bool__(self):
        return len(self) > 0 or len(self.objects) > 0

    @property
    def length(self):
        if self.last_response:
            return len(
                self.last_response["organization"]["orgEntity"]["artifactMemberships"][
                    "edges"
                ]
            )
        else:
            return None

    @property
    def more(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactMemberships"
            ]["pageInfo"]["hasNextPage"]
        else:
            return True

    @property
    def cursor(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactMemberships"
            ]["pageInfo"]["endCursor"]
        else:
            return None

    def convert_objects(self):
        if not self.last_response:
            return []
        if (
            not self.last_response["organization"]
            or not self.last_response["organization"]["orgEntity"]
        ):
            raise ValueError(
                f"Organization '{self.organization}' not found. Please verify the organization name is correct"
            )

        artifacts = (
            wandb.Artifact._from_attrs(
                a["node"]["artifactCollection"]["project"]["entity"]["name"],
                a["node"]["artifactCollection"]["project"]["name"],
                a["node"]["artifactCollection"]["name"]
                + ":v"
                + str(a["node"]["versionIndex"]),
                a["node"]["artifact"],
                self.client,
                [alias["alias"] for alias in a["node"]["aliases"]],
            )
            for a in self.last_response["organization"]["orgEntity"][
                "artifactMemberships"
            ]["edges"]
        )
        return artifacts


def _ensure_registry_prefix_on_names(query, in_name=False):
    """Traverse the filter to prepend the `name` key value with the registry prefix unless the value is a regex.

    - in_name: True if we are under a "name" key (or propagating from one).

    EX: {"name": "model"} -> {"name": "wandb-registry-model"}
    """
    if isinstance((txt := query), str):
        if in_name:
            return txt if txt.startswith(REGISTRY_PREFIX) else f"{REGISTRY_PREFIX}{txt}"
        return txt
    if isinstance((dct := query), Mapping):
        new_dict = {}
        for key, obj in dct.items():
            if key == "name":
                new_dict[key] = _ensure_registry_prefix_on_names(obj, in_name=True)
            elif key == "$regex":
                # For regex operator, we skip transformation of its value.
                new_dict[key] = obj
            else:
                # For any other key, propagate the in_name and skip_transform flags as-is.
                new_dict[key] = _ensure_registry_prefix_on_names(obj, in_name=in_name)
        return new_dict
    if isinstance((objs := query), Sequence):
        return list(
            map(lambda x: _ensure_registry_prefix_on_names(x, in_name=in_name), objs)
        )
    return query



"""Events that trigger W&B Automations."""

# ruff: noqa: UP007  # Avoid using `X | Y` for union fields, as this can cause issues with pydantic < 2.6

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from pydantic import Field
from typing_extensions import Self, TypeAlias, get_args

from wandb._pydantic import (
    GQLBase,
    SerializedToJson,
    field_validator,
    model_validator,
    pydantic_isinstance,
)

from ._filters import And, MongoLikeFilter, Or
from ._filters.expressions import FilterableField
from ._filters.run_metrics import (
    Agg,
    MetricChangeFilter,
    MetricOperand,
    MetricThresholdFilter,
)
from ._generated import EventTriggeringConditionType, FilterEventFields
from ._validators import simplify_op
from .actions import InputAction, InputActionTypes
from .scopes import ArtifactCollectionScope, InputScope, ProjectScope

if TYPE_CHECKING:
    from .automations import NewAutomation


# NOTE: Re-defined publicly with a more readable name for easier access
EventType = EventTriggeringConditionType
"""The type of event that triggers an automation."""

Agg = Agg


# ------------------------------------------------------------------------------
# Saved types: for parsing response data from saved automations


# Note: In GQL responses containing saved automation data, the filter is wrapped in an extra `filter` key.
class SavedEventFilter(GQLBase):
    filter: SerializedToJson[MongoLikeFilter] = Field(default_factory=And)


class _InnerRunMetricFilter(GQLBase):  # from `RunMetricFilter`
    threshold_filter: Optional[MetricThresholdFilter] = None
    change_filter: Optional[MetricChangeFilter] = None

    @model_validator(mode="before")
    @classmethod
    def _wrap_metric_filter(cls, v: Any) -> Any:
        if pydantic_isinstance(v, MetricThresholdFilter):
            return cls(threshold_filter=v)
        if pydantic_isinstance(v, MetricChangeFilter):
            return cls(change_filter=v)
        return v

    @model_validator(mode="after")
    def _ensure_exactly_one_set(self) -> Self:
        set_field_names = [name for name, val in self if (val is not None)]
        if not set_field_names:
            raise ValueError("Must specify a run metric filter")
        if len(set_field_names) > 1:
            names = ", ".join(map(repr, set_field_names))
            raise ValueError(f"Must specify a single run metric filter, got: {names}")
        return self


class RunMetricFilter(GQLBase):  # from `RunMetricEvent`
    run_filter: SerializedToJson[MongoLikeFilter] = Field(default_factory=And)
    run_metric_filter: _InnerRunMetricFilter

    #: Legacy field to define triggers on run metrics from absolute thresholds.  Use `run_metric_filter` instead.
    metric_filter: Optional[SerializedToJson[MetricThresholdFilter]] = Field(
        default=None,
        deprecated="The `metric_filter` field is deprecated: use `run_metric_filter` instead.",
    )

    @model_validator(mode="before")
    @classmethod
    def _wrap_metric_filter(cls, v: Any) -> Any:
        if pydantic_isinstance(v, (MetricThresholdFilter, MetricChangeFilter)):
            # If we're only given an (unwrapped) metric filter, automatically wrap it
            # in the appropriate nested structure, and use the default run filter.

            # Delegate to the inner validator to further wrap the filter as appropriate.
            return cls(run_metric_filter=_InnerRunMetricFilter.model_validate(v))
        return v

    @field_validator("run_filter", mode="after")
    @classmethod
    def _wrap_run_filter(cls, v: MongoLikeFilter) -> Any:
        v_new = simplify_op(v)
        return (
            And.model_validate(v_new)
            if pydantic_isinstance(v_new, And)
            else And(and_=[v_new])
        )


# type alias defined for naming consistency/clarity
SavedRunMetricFilter: TypeAlias = RunMetricFilter


class SavedEvent(FilterEventFields):  # from `FilterEventTriggeringCondition`
    """A more introspection-friendly representation of a triggering event from a saved automation."""

    # We override the type of the `filter` field since the original GraphQL
    # schema (and generated class) defines it as a JSONString (str), but we
    # have more specific expectations for the structure of the JSON data.
    filter: SerializedToJson[Union[SavedEventFilter, SavedRunMetricFilter]]


# ------------------------------------------------------------------------------
# Input types: for creating or updating automations


# Note: The GQL input for "eventFilter" does NOT wrap the filter in an extra `filter` key, unlike the
# eventFilter returned in responses for saved automations.
class _BaseEventInput(GQLBase):
    event_type: EventType
    scope: InputScope
    filter: SerializedToJson[Any]

    def add_action(self, action: InputAction) -> NewAutomation:
        """Define an executed action to be triggered by this event."""
        from .automations import NewAutomation

        if isinstance(action, InputActionTypes):
            return NewAutomation(scope=self.scope, event=self, action=action)

        raise TypeError(f"Expected a valid action, got: {type(action).__qualname__!r}")

    def __rshift__(self, other: InputAction) -> NewAutomation:
        """Supports `event >> action` as syntactic sugar to combine an event and action."""
        return self.add_action(other)


class _BaseMutationEventInput(_BaseEventInput):
    event_type: EventType
    scope: InputScope
    filter: SerializedToJson[MongoLikeFilter] = Field(default_factory=And)

    @field_validator("filter", mode="after")
    @classmethod
    def _wrap_filter(cls, v: Any) -> Any:
        """Ensure the given filter is wrapped like: `{"$or": [{"$and": [<original_filter>]}]}`.

        This is awkward but necessary, because the frontend expects this format.
        """
        v_new = simplify_op(v)
        v_new = (
            And.model_validate(v_new)
            if pydantic_isinstance(v_new, And)
            else And(and_=[v_new])
        )
        v_new = Or(or_=[v_new])
        return v_new


class OnLinkArtifact(_BaseMutationEventInput):
    """A new artifact is linked to a collection."""

    event_type: Literal[EventType.LINK_MODEL] = EventType.LINK_MODEL
    scope: InputScope


class OnAddArtifactAlias(_BaseMutationEventInput):
    """A new alias is assigned to an artifact."""

    event_type: Literal[EventType.ADD_ARTIFACT_ALIAS] = EventType.ADD_ARTIFACT_ALIAS
    scope: InputScope


class OnCreateArtifact(_BaseMutationEventInput):
    """A new artifact is created."""

    event_type: Literal[EventType.CREATE_ARTIFACT] = EventType.CREATE_ARTIFACT
    scope: ArtifactCollectionScope


class OnRunMetric(_BaseEventInput):
    """A run metric satisfies a user-defined absolute threshold."""

    event_type: Literal[EventType.RUN_METRIC] = EventType.RUN_METRIC
    scope: ProjectScope
    filter: SerializedToJson[RunMetricFilter]


# for type annotations
InputEvent = Union[
    OnLinkArtifact,
    OnAddArtifactAlias,
    OnCreateArtifact,
    OnRunMetric,
]
# for runtime type checks
InputEventTypes: tuple[type, ...] = get_args(InputEvent)


# ----------------------------------------------------------------------------


class RunEvent:
    name = FilterableField("display_name")
    # `Run.name` is actually filtered on `Run.display_name` in the backend.
    # We can't reasonably expect users to know this a priori, so
    # automatically fix it here.

    @staticmethod
    def metric(name: str) -> MetricOperand:
        """Define a metric filter condition."""
        return MetricOperand(name=name)


class ArtifactEvent:
    alias = FilterableField()


MetricThresholdFilter.model_rebuild()
RunMetricFilter.model_rebuild()
SavedEventFilter.model_rebuild()

OnLinkArtifact.model_rebuild()
OnAddArtifactAlias.model_rebuild()
OnCreateArtifact.model_rebuild()
OnRunMetric.model_rebuild()



from typing import Union

from pydantic import Field

from wandb._pydantic import GQLBase
from wandb.automations._generated import (
    GenericWebhookIntegrationFields,
    SlackIntegrationFields,
)


class SlackIntegration(SlackIntegrationFields):
    pass


class WebhookIntegration(GenericWebhookIntegrationFields):
    pass


Integration = Union[SlackIntegration, WebhookIntegration]


# For parsing integration instances from paginated responses
class _IntegrationEdge(GQLBase):
    cursor: str
    node: Integration = Field(discriminator="typename__")



# ruff: noqa: UP007  # Avoid using `X | Y` for union fields, as this can cause issues with pydantic < 2.6

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional

from pydantic import Field

from wandb._pydantic import Base, GQLId

from ._generated import TriggerFields, UserFields
from .actions import InputAction, SavedAction
from .events import InputEvent, SavedEvent
from .scopes import InputScope, SavedScope

if TYPE_CHECKING:
    pass


# ------------------------------------------------------------------------------
# Saved types: for parsing response data from saved automations
class Automation(TriggerFields):
    """A local instance of a saved W&B automation."""

    id: GQLId

    created_by: UserFields = Field(repr=False, frozen=True, alias="createdBy")
    created_at: datetime = Field(repr=False, frozen=True, alias="createdAt")
    updated_at: Optional[datetime] = Field(repr=False, frozen=True, alias="updatedAt")

    name: str
    description: Optional[str]

    scope: SavedScope = Field(discriminator="typename__")
    event: SavedEvent
    action: SavedAction = Field(discriminator="typename__", alias="triggeredAction")

    enabled: bool

    # def save(
    #     self, api: Api | None = None, **updates: Unpack[AutomationParams]
    # ) -> Automation:
    #     """Save this existing automation to the server, applying any local changes.

    #     Args:
    #         api: The API instance to use.  If not provided, the default API instance is used.
    #         updates:
    #             Any final updates to apply to the automation before
    #             saving it.  These override previously-set values, if any.

    #     Returns:
    #         The updated automation.
    #     """
    #     from wandb import Api

    #     return (api or Api()).update_automation(self, **updates)

    # def delete(self, api: Api | None = None) -> DeleteTriggerResult:
    #     """Delete this automation from the server.

    #     Args:
    #         api: The API instance to use.  If not provided, the default API instance is used.
    #     """
    #     from wandb import Api

    #     return (api or Api()).delete_automation(self)


class NewAutomation(Base):
    """An automation which can hold any of the fields of a NewAutomation, but may not be complete yet."""

    name: Optional[str] = None
    description: Optional[str] = None
    enabled: bool = True

    scope: Optional[InputScope] = Field(discriminator="typename__", default=None)
    event: Optional[InputEvent] = Field(discriminator="event_type", default=None)
    action: Optional[InputAction] = Field(discriminator="action_type", default=None)

    # def save(
    #     self, api: Api | None = None, **updates: Unpack[AutomationParams]
    # ) -> Automation:
    #     """Create this automation by saving it to the server.

    #     Args:
    #         api: The API instance to use.  If not provided, the default API instance is used.
    #         updates:
    #             Any final updates to apply to the automation before
    #             saving it.  These override previously-set values, if any.

    #     Returns:
    #         The created automation.
    #     """
    #     from wandb import Api

    #     return (api or Api()).create_automation(self, **updates)


class PreparedAutomation(NewAutomation):
    """A fully defined automation, ready to be sent to the server to create or update it."""

    name: str
    description: Optional[str] = None
    enabled: bool = True

    scope: InputScope
    event: InputEvent
    action: InputAction



"""Pydantic-compatible representations of MongoDB expressions."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Union

from pydantic import ConfigDict, model_serializer
from typing_extensions import Self, TypeAlias

from wandb._pydantic import CompatBaseModel, model_validator

from .operators import (
    Contains,
    Eq,
    Exists,
    Gt,
    Gte,
    In,
    Lt,
    Lte,
    Ne,
    NotIn,
    Op,
    Regex,
    RichReprResult,
    Scalar,
    ScalarTypes,
    SupportsLogicalOpSyntax,
)


class FilterableField:
    """A descriptor that can be used to define a "filterable" field on a class.

    Internal helper to support syntactic sugar for defining event filters.
    """

    _python_name: str  #: The name of the field this descriptor was assigned to in the Python class.
    _server_name: str | None  #: If set, the actual server-side field name to filter on.

    def __init__(self, server_name: str | None = None):
        self._server_name = server_name

    def __set_name__(self, owner: type, name: str) -> None:
        self._python_name = name

    def __get__(self, obj: Any, objtype: type) -> Self:
        # By default, if we didn't explicitly provide a backend name for
        # filtering, assume the field has the same name in the backend as
        # the python attribute.
        return self

    @property
    def _name(self) -> str:
        return self._server_name or self._python_name

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._name!r})"

    # Methods to define filter expressions through chaining
    def matches_regex(self, pattern: str) -> FilterExpr:
        return FilterExpr(field=self._name, op=Regex(regex_=pattern))

    def contains(self, text: str) -> FilterExpr:
        return FilterExpr(field=self._name, op=Contains(contains_=text))

    def exists(self, exists: bool = True) -> FilterExpr:
        return FilterExpr(field=self._name, op=Exists(exists_=exists))

    def lt(self, value: Scalar) -> FilterExpr:
        return FilterExpr(field=self._name, op=Lt(lt_=value))

    def gt(self, value: Scalar) -> FilterExpr:
        return FilterExpr(field=self._name, op=Gt(gt_=value))

    def lte(self, value: Scalar) -> FilterExpr:
        return FilterExpr(field=self._name, op=Lte(lte_=value))

    def gte(self, value: Scalar) -> FilterExpr:
        return FilterExpr(field=self._name, op=Gte(gte_=value))

    def ne(self, value: Scalar) -> FilterExpr:
        return FilterExpr(field=self._name, op=Ne(ne_=value))

    def eq(self, value: Scalar) -> FilterExpr:
        return FilterExpr(field=self._name, op=Eq(eq_=value))

    def in_(self, values: Iterable[Scalar]) -> FilterExpr:
        return FilterExpr(field=self._name, op=In(in_=values))

    def not_in(self, values: Iterable[Scalar]) -> FilterExpr:
        return FilterExpr(field=self._name, op=NotIn(nin_=values))

    # Override the default behavior of comparison operators: <, >=, ==, etc
    def __lt__(self, other: Any) -> FilterExpr:
        if isinstance(other, ScalarTypes):
            return self.lt(other)
        raise TypeError(f"Invalid operand type in filter expression: {type(other)!r}")

    def __gt__(self, other: Any) -> FilterExpr:
        if isinstance(other, ScalarTypes):
            return self.gt(other)
        raise TypeError(f"Invalid operand type in filter expression: {type(other)!r}")

    def __le__(self, other: Any) -> FilterExpr:
        if isinstance(other, ScalarTypes):
            return self.lte(other)
        raise TypeError(f"Invalid operand type in filter expression: {type(other)!r}")

    def __ge__(self, other: Any) -> FilterExpr:
        if isinstance(other, ScalarTypes):
            return self.gte(other)
        raise TypeError(f"Invalid operand type in filter expression: {type(other)!r}")

    # Operator behavior is intentionally overridden to allow defining
    # filter expressions like `field == "value"`.  See similar overrides
    # of built-in dunder methods in sqlalchemy, polars, pandas, numpy, etc.
    #
    # sqlalchemy example for illustrative purposes:
    # https://github.com/sqlalchemy/sqlalchemy/blob/f21ae633486380a26dc0b67b70ae1c0efc6b4dc4/lib/sqlalchemy/orm/descriptor_props.py#L808-L812
    def __eq__(self, other: Any) -> FilterExpr:
        if isinstance(other, ScalarTypes):
            return self.eq(other)
        raise TypeError(f"Invalid operand type in filter expression: {type(other)!r}")

    def __ne__(self, other: Any) -> FilterExpr:
        if isinstance(other, ScalarTypes):
            return self.ne(other)
        raise TypeError(f"Invalid operand type in filter expression: {type(other)!r}")


# ------------------------------------------------------------------------------
class FilterExpr(CompatBaseModel, SupportsLogicalOpSyntax):
    """A MongoDB filter expression on a specific field."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    field: str
    op: Op

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.field!s}={self.op!r})"

    def __rich_repr__(self) -> RichReprResult:  # type: ignore[override]
        # https://rich.readthedocs.io/en/stable/pretty.html
        yield self.field, self.op

    @model_validator(mode="before")
    @classmethod
    def _validate(cls, data: Any) -> Any:
        """Parse a MongoDB dict representation of the filter expression."""
        if (
            isinstance(data, dict)
            and len(data) == 1
            and not any(key.startswith("$") for key in data)
        ):
            # This looks like a MongoDB filter dict.  E.g.:
            # - in:  `{"display_name": {"$contains": "my-run"}}`
            # - out: `FilterExpr(field="display_name", op=Contains(contains_="my-run"))`
            ((field, op),) = data.items()
            return {"field": field, "op": op}
        return data

    @model_serializer(mode="plain")
    def _serialize(self) -> dict[str, Any]:
        """Return a MongoDB dict representation of the expression."""
        from pydantic_core import to_jsonable_python  # Only valid in pydantic v2

        op_dict = to_jsonable_python(self.op, by_alias=True, round_trip=True)
        return {self.field: op_dict}


MongoLikeFilter: TypeAlias = Union[Op, FilterExpr]



from __future__ import annotations

from dataclasses import dataclass
from typing import Union
from urllib import parse


@dataclass
class RunMoment:
    """A moment in a run.

    Defines a branching point in a finished run to fork or resume from.
    A run moment is identified by a run ID and a metric value.
    Currently, only the metric '_step' is supported.
    """

    run: str
    """run ID"""

    value: Union[int, float]
    """Value of the metric."""

    metric: str = "_step"
    """Metric to use to determine the moment in the run.

    Currently, only the metric '_step' is supported.
    In future, this will be relaxed to be any metric.
    """

    def __post_init__(self):
        if self.metric != "_step":
            raise ValueError(
                f"Only the metric '_step' is supported, got '{self.metric}'."
            )
        if not isinstance(self.value, (int, float)):
            raise ValueError(
                f"Only int or float values are supported, got '{self.value}'."
            )
        if not isinstance(self.run, str):
            raise ValueError(f"Only string run names are supported, got '{self.run}'.")

    @classmethod
    def from_uri(cls, uri: str) -> RunMoment:
        parsable = "runmoment://" + uri
        parse_err = ValueError(
            f"Could not parse passed run moment string '{uri}', "
            f"expected format '<run>?<metric>=<numeric_value>'. "
            f"Currently, only the metric '_step' is supported. "
            f"Example: 'ans3bsax?_step=123'."
        )

        try:
            parsed = parse.urlparse(parsable)
        except ValueError as e:
            raise parse_err from e

        if parsed.scheme != "runmoment":
            raise parse_err

        # extract run, metric, value from parsed
        if not parsed.netloc:
            raise parse_err

        run = parsed.netloc

        if parsed.path or parsed.params or parsed.fragment:
            raise parse_err

        query = parse.parse_qs(parsed.query)
        if len(query) != 1:
            raise parse_err

        metric = list(query.keys())[0]
        if metric != "_step":
            raise parse_err
        value: str = query[metric][0]
        try:
            num_value = int(value) if value.isdigit() else float(value)
        except ValueError as e:
            raise parse_err from e

        return cls(run=run, metric=metric, value=num_value)
